import re

from ..lib import Invalid

from .core import ValidatorBase, Validator, Compose, Tmp, Item, Call, If, messages
from .basic import String, Dict, DateTime
from .alter import Encode, Decode, Lower, EliminateWhiteSpace, Split, Join, UpdateValue
from .check import Match, In, Len
from .schema import Schema, ForEach

from . import cache

@messages\
    ( fail='Domain offers no mailserver'
    )
class MXLookup( Validator ):
    def on_value( self, context, value ):
        import dns.resolver
        r = None
        try:
            r = dns.resolver.query(value, 'MX')
        except dns.resolver.NXDOMAIN:
            pass
        if not r:
            raise Invalid( value, self )

        return value


CommonDomainPreValidaton =\
    ( String.convert().tag('string')\
    & EliminateWhiteSpace().tag('eliminateWhiteSpace')\
    & Lower().tag('toLower')\
    & UpdateValue().tag('update')
    ).tag('prevalidation')

ComposedDomainLabel = Compose\
    ( CommonDomainPreValidaton
    & cache.Set('domainLabel')
    &   If  ( Match( re.compile(r'^xn--') )
            , Tmp( Encode('utf8') & Decode('idna').tag('decodeIdna') & cache.Set('domainLabelUnicode') )
            , cache.Set('domainLabelUnicode') & Encode('idna').tag('encodeIdna') & Decode('utf8')
            ).tag('idna')
    & cache.Set('domainLabel')
    & Len(max=63).tag('length')
    & Match(re.compile(r'^((([a-z][0-9])|([0-9][a-z])|([a-z0-9][a-z0-9\-]{1,2}[a-z0-9])|([a-z0-9][a-z0-9\-](([a-z0-9\-][a-z0-9])|([a-z0-9][a-z0-9\-]))[a-z0-9\-]*[a-z0-9]))|([a-z0-9]{1,2})|(xn\-\-[\-a-z0-9]*[a-z0-9]))$')).tag('validSymbols')
    #& create.List( cache.Get('domainLabel'), cache.Get('domainLabelUnicode') )
    & cache.Get('domainLabelUnicode').tag('returnUnicode')
    ).paramAlias\
        ( convertToString='string_convert'
        , updateValue='update_enabled'
        , eliminateWhiteSpace='eliminateWhiteSpace_enabled'
        , toLower='toLower_enabled'
        , extractIdna='idna_enabled'
        , returnUnicode='returnUnicode_enabled'
    ).messageAlias\
        ( type='string_type'
        , tooLong='length_max'
        , invalidSymbols=('validSymbols_fail','encodeIdna_fail','decodeIdna_fail')
        , blank=
            ("toLower_blank"
            ,"string_blank"
            ,"encodeIdna_blank"
            ,"decodeIdna_blank"
            ,"length_blank"
            )
        , missing="string_missing"
    ).messages\
        ( blank='Please provide a value'
        , tooLong='A domain label can have max %(max)i characters'
        , invalidSymbols='The domain name contains invalid symbols'
        )


def __restrictToTLDSetter( alias, param ):
    if isinstance( param, list ) or isinstance( param, tuple ) or isinstance( param, set ):
        return\
            { 'restrictToTLDValidator_enabled': True
            , 'restrictToTLD_criteria': param
            }
    else:
        return\
            { 'restrictToTLDValidator_enabled': False
            }

def __domain_save_nonidna( context, value ):
    domainName = context.cache.get('domainName',None)
    
    if domainName is None:
        domainName = ''
    else:
        domainName += '.'

    domainName += context.cache.get('domainLabel')
    context.cache['domainName'] = domainName
    return value

Domain = Compose\
    ( CommonDomainPreValidaton
    & Split('.').tag('split')
    & Len(min=2).tag('numSubdomains')
    & ForEach\
        ( ComposedDomainLabel\
            ( prevalidation_enabled=False
            ).tag('domainLabel') & Call(__domain_save_nonidna)
        , createContextChildren=False
        )
    & Item\
        ( -1
        , In([]).tag('restrictToTLD')
        , alter=False
        ).tag('restrictToTLDValidator', False)
    & Join('.')\
    & Tmp( cache.Get('domainName') & MXLookup().tag('mxLookup') ).tag('resolve',False)\
    ).paramAlias\
        ( convertToString='string_convert'
        , extractIdna='domainLabel_extractIdna'
        , returnUnicode='domainLabel_returnUnicode'
        , updateValue='update_enabled'
        , eliminateWhiteSpace='eliminateWhiteSpace_enabled'
        , toLower='toLower_enabled'
        , resolve='resolve_enabled'
        , restrictToTLD= __restrictToTLDSetter
    ).messageAlias\
        ( blank=('string_blank','toLower_blank')
        , missing='string_missing'
        , tooLong='domainLabel_tooLong'
        , type='string_type'
        , format = ('numSubdomains_min','domainLabel_blank')
        , restrictToTLD= 'restrictToTLD_fail'
        , invalidSymbols='domainLabel_invalidSymbols'
        , resolve='mxLookup_fail'
    ).messages\
        ( blank="Please provide a value"
        , format='Invalid domain name format, try my.domain.com'
        , restrictToTLD= 'TLD not allowed. Allowed TLDs are %(criteria)s'
        , tooLong="A domain label cannot exceed %(max)i characters"
        )


EmailLocalPart = Compose\
    (   ( String.convert().tag('string')
        & EliminateWhiteSpace().tag('eliminateWhiteSpace')
        & UpdateValue().tag('update')
        ).tag('prevalidation')
    & Len(max=64).tag('length')
    & Match(re.compile(r'^[a-z0-9!#$%&\'\*\+\-\/\=\?\^_`\{\|\}~]+(\.[a-z0-9!#$%&\'\*\+\-\/\=\?\^_`\{\|\}~]+)*$', re.I)).tag('validSymbols')
    ).paramAlias\
        ( convertToString='string_convert'
        , updateValue='update_enabled'
        , eliminateWhiteSpace='eliminateWhiteSpace_enabled'
    ).messageAlias\
        ( blank=('string_blank','validSymbols_blank','length_blank')
        , missing='string_missing'
        , tooLong='length_max'
        , type='string_type'
        , invalidSymbols='validSymbols_fail'
    ).messages\
        ( tooLong='Email local-part may only be up to %(max)i characters long'
        , invalidSymbols='Localpart contains invalid symbols'
        )


Email = Compose\
    ( String.convert().tag('string')
    & EliminateWhiteSpace().tag('eliminateWhiteSpace')
    & Split('@',1).tag('split')
    & Tmp\
        ( Item( 1 , Lower().tag('doLowerDomainPart') ).tag('itemDomainPart')
        & Join('@' )
        & UpdateValue().tag('update')
        ).tag('lowerDomainPart')
    & Schema\
        ( 'localPart'
            ,  cache.Set('localPart') & EmailLocalPart( prevalidation_enabled=False ).tag('localPart')
        , 'domainPart'
            ,  cache.Set('domainPart') & Domain( prevalidation_enabled=False ).tag('domainPart')
        , createContextChildren=False
        , returnList=True
        )
    & Join('@')
    ).paramAlias\
        ( eliminateWhiteSpace = 'eliminateWhiteSpace_enabled'
        , updateValue = 'update_enabled'
        , lowerDomainPart = 'lowerDomainPart_enabled'
    ).messageAlias\
        ( blank = ('string_blank','split_blank' )
        , missing = 'string_missing'
        , format = \
            ( 'itemDomainPart_blank'
            , 'itemDomainPart_notFound'
            , 'doLowerDomainPart_blank'
            , 'localPart_blank'
            , 'domainPart_blank'
            )
    ).messages\
        ( blank = 'Please enter an email address'
        , format = 'Invalid email format ( try my.email@address.com )'
        , localPart_invalidSymbols = "The part before @ (%(localPart)s) contains invalid symbols"
        , domainPart_restrictToTLD="Invalid top level domain %(domainLabel)s, allowed TLD are %(criteria)s"
        , domainPart_tooLong="Domain part is too long. Max %(max)s characters allowed per domain label"
        , domainPart_format="Invalid domain name format: %(domainPart)s"
        , domainPart_invalidSymbols="Domain part contains invalid characters: %(domainLabel)s"
        )


    
DateField = Compose\
    ( String.convert()
    & EliminateWhiteSpace()
    &   ( DateTime.convert( '%y-%m-%d' )
        | DateTime.convert( '%Y-%m-%d' )
        | DateTime.convert( '%d.%m.%y' )
        | DateTime.convert( '%d.%m.%Y' ).tag('dateTimeConverter')
        )
    ).messageAlias\
        ( format='dateTimeConverter_convert'
    ).messages\
        ( format='Invalid date format ( try YY(YY)-MM-DD or DD.MM.YY(YY) )'
        )


class NestedPostConverter( ValidatorBase ):


    def validate( self, context, value ):
        resultset = {}

        for (key, val) in value.items():
            parts = key.split('.')

            result = resultset

            while len(parts)>1:
                part = parts.pop(0)
                if not part in result:
                    result[part] = {}
                result = result[part]
 
            result[parts[0]] = val

        return resultset


NestedPost = Compose\
    ( Dict().tag('type')
    & NestedPostConverter()
    ).messageAlias\
        ( type='type_fail'
        )



from kanone import Invalid
from decorator import decorator

def validate( validator, errorFormatter=None ):

    def __wrapper__( origFunction  ):

        def __formFunction__( f, self, *args, **kwargs ):
            request = self._py_object.request
            self.form_context = context = validator.context( dict(request.params) )

            if request.params:
                if errorFormatter is not None:
                    context.errorFormatter = errorFormatter
                try:
                    result = context.result
                except Invalid:
                    pass
                else:
                    return __formFunction__.__successFunc__( self,  result, *args, **kwargs )
            else:
                __formFunction__.__initFunc__( self, context, *args, **kwargs )
        
            return f( self, *args, **kwargs )
       
        resultFunc = decorator( __formFunction__, origFunction )

        def initFormFunction(  initFunc ):
            __formFunction__.__initFunc__ = initFunc
        
            return resultFunc
        
        def successFormFunction(  successFunc ):
            __formFunction__.__successFunc__ = successFunc
        
            return resultFunc

        resultFunc.init = initFormFunction
        resultFunc.success = successFormFunction
        return resultFunc

    return __wrapper__



from ..error import Invalid
from ..util import varargs2kwargs, getArgSpec, getParameterNames

import logging

log = logging.getLogger(__name__)

def validateDecorator( validator, method, include, exclude, onInvalid ):

    if include and exclude:
        raise SyntaxError("'include' and 'exclude' cannot be used at the same time")

    spec = getArgSpec( method )
    hasVarargs = spec.varargs is not None
    varargs =  spec.varargs or '*varargs'
    keywords = spec.keywords or False

    methodParameterNames = getParameterNames( method, skipSelf=False )

    skip = ()
    if exclude:
        skip = exclude
    if include:
        skip = set(methodParameterNames) - set(include)

    varargs    = varargs

    hasVarargs = spec.varargs not in skip and hasVarargs

    keywords   = keywords not in skip and keywords

    def __wrap( *fargs, **fkwargs):

        (fargs, fkwargs, shifted ) = varargs2kwargs( method, fargs, fkwargs, skipSelf=False )
        origKwargs = dict(fkwargs)

        if keywords is not False:
            restKwargs = dict(\
                ( key, fkwargs.pop(key))\
                    for key in list(fkwargs.keys()) if key not in methodParameterNames
                )
            fkwargs[ keywords ] = restKwargs

        if fargs or hasVarargs:
            fkwargs[ varargs ] = list(fargs)

        try:
            resultKwargs =validator.context\
                ( dict( ( key, fkwargs[ key] ) for key in fkwargs if key not in skip ) ).result
        except Invalid as e:
            if onInvalid is not None:
                return onInvalid( e )
            else:
                raise

        origKwargs.update( resultKwargs )

        resultArgs = origKwargs.pop( varargs, fargs )
        resultArgs = [ origKwargs.pop(key) for key in shifted  ] + resultArgs

        if keywords is not False:
            origKwargs.update( origKwargs.pop( keywords ) )

        return method( *resultArgs, **origKwargs )

    return __wrap

def validate( validator, include=None, exclude=None, onInvalid=None ):
    def __createDecorator( method ):
        return validateDecorator( validator, method, include, exclude, onInvalid)
    return __createDecorator




""" Twisted adapter for Kanone """

from twisted.python.failure import Failure
from twisted.internet import defer
from ..lib import Invalid
from ..util import varargs2kwargs

import logging, sys
log = logging.getLogger( __name__ )

# hacky and redundant, but it'll do for now ..
# TODO: move to proper twisted specific classes under .tx.*
#       and get rid of the monkey

_python3 = sys.version_info[0]>=3

def monkeyPatch():
    """
    Patches Kanone so that any validation returns a Deferred, thus
    one can write asynchronous validators using Twisted's non-blocking API.
    Schema and ForEach fields are validated concurrently.
    """
    if getattr( monkeyPatch,'_isMonkeyPatched',False):
        return

    from ..lib import Context, PASS, MISSING

    from ..validator.core import Tag, Compose, Tmp, Item, Not, And, Or, Call, If
    from ..validator.check import Match
    from ..validator.schema import Schema, ForEach, Field
    from ..validator.web import MXLookup

    @defer.inlineCallbacks
    def context_validate( self ):
        if self.isValidated:
            if self.__error__ is not MISSING:
                raise self.__error__

            defer.returnValue( self.__result__ )

        self.isValidating = True

        if self.parent is not None:

            if not self.parent.isValidated and not self.parent.isValidating:
                yield defer.maybeDeferred\
                    ( self.parent.validate
                    )

        if not self.validator:
            raise AttributeError("No validator set for context '%s'" % self.path )

        result = defer.maybeDeferred\
            ( self.validator.validate
            , self
            , self.__value__
            )

        result.addErrback( context_gotError, self )
        result = yield result

        self.isValidated = True
        self.isValidating = False

        if self.__error__ is not MISSING:
            raise self.__error__
        else:
            if result is not PASS:
                self.__result__ = result
            else:
                self.__result__ = self.__value__

            self.__result__ = result

        defer.returnValue( result )


    def context_gotError( error, self ):

        e = error.value 

        if not isinstance( e, Invalid ):
            self.__error__ = error
            return

        self.__error__ = e
        e.context = self
        message = e.validator.__messages__[e.key]

        if message is not None:
            extra = e.data['extra']
            value = e.value
            data = e.data

            data['message'] = message
            if hasattr(e,'realkey'):
                data['key'] = e.realkey

            extra['value.type'] = getattr(value, '__class__', None) is not None \
                and getattr(value.__class__,'__name__', False) or 'unknown'

            if isinstance(value,str) or not _python3 and isinstance(value,basestring):
                extra['value'] = value
            else:
                extra['value'] = str(value)

            cache = getattr( self, 'cache', None)
            if cache is not None:
                extra.update( cache )

            self['error'] = self.__error__.data

            self.root.errorlist.append( self.__error__.context.path )


    def tag_gotResult( result, d, validator, tagName ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                d.errback( result )
                return

            e = result.value
            if e.validator is validator or getattr(e,'composer',None) is validator:
                e.tagName = tagName
            d.errback( e )
        else:
            d.callback( result )


    def tag_validate( self, context, value ):
        validator = context.root.taggedValidators.get(self.tagID, None)
        if validator is None:
            validator = self.enabled and self.validator

        if not validator:
            return value

        d = defer.Deferred()
        result = defer.maybeDeferred\
                ( validator.validate
                , context
                , value
                )
        result.addBoth( tag_gotResult, d, validator, self.tagName )
        return d

    def compose_gotResult( result, d, context, tmpTags, composer ):
        context.root.taggedValidators = tmpTags

        if isinstance( result, Failure ):
            if not isinstance( result.value, Invalid ):
                d.errback( result )
                return

            e = result.value

            if hasattr(e,'tagName'):
                e.realkey = "%s_%s" % (e.tagName, getattr(e,'realkey',e.key))
                e.composer = composer
                del e.tagName
            d.errback( e )
        else:
            d.callback( result )



    def compose_validate( self, context, value ):
        tmpTags = context.root.taggedValidators
        context.root.taggedValidators = self.currentTaggedValidators

        d = defer.Deferred()
        result = defer.maybeDeferred\
            ( self.validator.validate
            , context
            , value
            )
        result.addBoth( compose_gotResult, d, context, tmpTags, self )
        return d

    def tmp_gotReslt( result, d, raiseError, value ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                d.errback( result )
                return

            if raiseError:
                d.errback( result.value )
                return

        d.callback( value )

    def tmp_validate( self, context, value ):
        d = defer.Deferred()
        result = defer.maybeDeferred\
                ( self.validator.validate
                , context
                , value
                )
        result.addBoth( tmp_gotReslt, d, self.raiseError, value )
        return d

    def item_gotResult( result, d, value, key, alter ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                d.errback( result )
                return
            d.errback( result.value )
        else:
            if alter:
                value[key] = result

            d.callback( value )

    def item_validate( self, context, value ):
        try:
            val = value[ self.key ]

        except TypeError:
            raise Invalid( value, self, 'type' )
        except (KeyError, IndexError):
            raise Invalid( value, self, 'notFound', key=self.key )
        else:
            if self.validator is not None:
                d = defer.Deferred()
                result = defer.maybeDeferred\
                    ( self.validator.validate
                    , context
                    , val
                    )
                result.addBoth( item_gotResult, d , value, self.key, self.alter )
                return d
            else:
                return val

    def not_gotResult( result, d, value, validator ):
        if isinstance( result, Failure ):
            if not isinstance( result.value, Invalid ):
                d.errback( result )
                return
            d.callback( value )
        else:
            d.errback( Invalid( value, validator ) )

    def not_validate(self, context, value ):
        d = defer.Deferred()
        result = defer.maybeDeferred\
            ( self.validator.validate
            , context
            , value
            )
        result.addBoth( not_gotResult, d, value, self )
        return d

    def and_doTryNext( result, validators, context, value, d ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                d.errback( result )
            else:
                e = result.value
                d.errback( e )
        else:
            if validators:
                and_tryNext( validators, context, result, d )
            else:
                d.callback( result )

    def and_tryNext( validators, context, value, d ):
        result = defer.maybeDeferred\
            ( validators.pop(0).validate
            , context
            , value
            )

        result.addBoth( and_doTryNext, validators, context, value, d )

    def and_validate( self, context, value ):
        d = defer.Deferred()
        and_tryNext( list( self.validators ), context, value, d )
        return d

    def or_doTryNext( result, validators, context, value, d ):
        if isinstance( result, Failure ):
            err = result
        
            if not isinstance(err.value, Invalid):
                d.errback( err )
                return

            e = err.value
            if not validators:
                d.errback( e )
            else:
                or_tryNext( validators, context, value, d )
        else:
            d.callback( result )

    def or_tryNext( validators, context, value, d ):
        result = defer.maybeDeferred\
            ( validators.pop(0).validate
            , context
            , value
            )

        result.addBoth( or_doTryNext, validators, context, value, d )

    def or_validate( self, context, value ):
        d = defer.Deferred()
        or_tryNext( list(self.validators), context, value, d )
        return d



    @defer.inlineCallbacks
    def call_validate( self, context, value ):
        try:
            result = yield defer.maybeDeferred\
                ( self.__func__
                , context
                , value
                )
        except Failure as e:
            if not isinstance(e.value, Invalid):
                raise
            e = e.value
            e.validator = self
            raise e
        else:
            defer.returnValue( result )

    def match_gotResult( result, self, value, d ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                raise

            d.errback( Invalid( value, self, matchType=self.type, criterion=result.value ) )
        else:
            val = value
            if self.ignoreCase:
                result = str(result).lower()
                val = str(value).lower()

            if val != result:
                d.errback( Invalid( value, self, matchType=self.type, criterion=result ) )
            else:
                d.callback( value )
           

    def match_on_value(self, context, value ):
        if self.type is Match.REGEX:
            if not self.criterion.match(value):
                raise Invalid( value, self, matchType=self.type, criterion=self.criterion.pattern)
            return value
        elif self.type is Match.VALIDATOR:
            compare = defer.maybeDeferred\
                ( self.criterion.validate
                , context
                , value
                )

            d = defer.Deferred()
            compare.addBoth( match_gotResult, self, value, d )
            return d
        else:
            compare = self.criterion

        val = value
        if self.ignoreCase:
            compare = str(compare).lower()
            val = str(value).lower()

        if val != compare:
            raise Invalid( value, self, matchType=self.type, criterion=compare )

        return value


    def if_gotResult( result, d, context, value ):
        if isinstance( result, Failure ):
            if not isinstance(result.value, Invalid):
                d.errback( result )
            else:
                d.errback( result.value )
        else:
            d.callback( result )

    def if_gotResultExpression( result, validator, d, context, value ):
        if isinstance( result, Failure ):
            if not isinstance( result.value, Invalid):
                raise
            value = defer.maybeDeferred\
                ( validator._else.validate, context, value
                )
        else:
            value = defer.maybeDeferred\
                ( validator._then.validate, context, result
                )

        value.addBoth( if_gotResult, d, context, value )

    def if_validate( self, context, value ):
        d = defer.Deferred()
        result = defer.maybeDeferred( self.criterion.validate, context, value )
        result.addBoth( if_gotResultExpression, self, d, context, value )
        return d

    def schema_gotResult( result, resultset, key, isList, returnList ):
        if returnList:
            resultset.append( result )
        else:
            resultset[ key ] = result

        return result

    def schema_gotError( error, errorset, key ):
        if isinstance( error, Failure ):
            if not isinstance(error.value, Invalid):
                raise error

            error = error.value

        errorset.append( error )

    def schema__on_value_done( waste, d, schema, value, result, errors ):
        if not errors:
            d.callback( result )
        else:
            d.errback( errors.pop(0) )

    def schema__createContextChildren_on_value_done( waste, d, schema, value, result, errors ):
        if not errors:
            d.callback( result )
        else:
            d.errback( Invalid( value, schema ) )

    def schema__on_value( self, context, value ):
        isList = isinstance(value, list) or isinstance(value,tuple) or isinstance(value,set)
        if not isList and not isinstance( value, dict ):
            raise Invalid( value, self, 'type')

        extraFields = None
        if not self.allowExtraFields:
            if isList:
                extraFields = max( len(value), len(self.index) )
            else:
                extraFields = list(value.keys())

        if self.returnList:
            result = []
        else:
            result = {}

        numValues = len(value)
        jobs = []

        errorset = []
        for pos in range(len(self.index)):
            key = self.index[pos]
            if isList:
                if numValues>pos:
                    val = value[ pos ]
                    if not self.allowExtraFields:
                        extraFields-=1
                else:
                    val = MISSING
            else:
                val = value.get( key, MISSING)
                if not self.allowExtraFields and val is not MISSING:
                    try: extraFields.remove(key)
                    except: pass

            job = defer.maybeDeferred\
                ( self.validators[ key ].validate
                , context
                , val
                )

            jobs.append\
                ( job.addCallback( schema_gotResult, result, key, isList, self.returnList )\
                    .addErrback( schema_gotError, errorset, key )
                )

        if extraFields:
            raise Invalid( value, self, 'extraFields',extraFields=extraFields)

        d = defer.Deferred()
        jobs =defer.DeferredList( jobs )
        jobs.addCallback\
            ( schema__on_value_done
            , d
            , self
            , value
            , result
            , errorset
            )

        return d

    def schema__createContextChildren_on_value( self, context, value ):
        isList = isinstance(value, list) or isinstance(value,tuple) or isinstance(value,set)

        if not isList and not isinstance( value, dict ):
            raise Invalid( value, self, 'type')

        extraFields = None
        if not self.allowExtraFields:
            if isList:
                extraFields = max( len(value), len(self.index) )
            else:
                extraFields = list(value.keys())

        errors = []

        if self.returnList:
            result = []
        else:
            result = {}

        len_value = len(value)
        len_index = len(self.index)

        # populate
        for pos in range(len_index):
            key = self.index[pos]
            childContext = context( key )
            try:
                childContext.validator = self.validators[ key ]
            except KeyError:
                raise SyntaxError("No validator set for %s" % childContext.path)

            if isList:
                if len_value<=pos:
                    childContext.__value__ = MISSING
                else:
                    childContext.__value__ = value[ pos ]
            else:
                childContext.__value__ = value.get( key, MISSING )

            if not self.allowExtraFields:
                if isList:
                    extraFields-=1
                else:
                    try: extraFields.remove(key)
                    except: pass

        if extraFields:
            raise Invalid( value, self, 'extraFields',extraFields=extraFields)

        context.setIndexFunc( lambda index: self.index[index] )

        jobs = []

        # validate
        for key in self.index:

            jobs.append\
                ( context( key ).result\
                    .addCallback( schema_gotResult, result, key, isList, self.returnList )\
                    .addErrback( schema_gotError, errors, key )
                )

        d = defer.Deferred()
        jobs = defer.DeferredList( jobs )
        jobs.addCallback\
            ( schema__createContextChildren_on_value_done
            , d
            , self
            , value
            , result
            , errors
            )

        return d

    def forEach__on_value( self, context, value ):
        if self.returnList:
            result = []
        else:
            result = {}

        isList = isinstance( value, list) or isinstance(value, tuple) or isinstance(value, set)
        errorset = []
        jobs = []
        if isList or self.numericKeys:
            for pos in range( len( value ) ):
                if not isList:
                    val = value.get(str(pos),MISSING)
                    if val is MISSING:
                        raise Invalid( value, self, 'numericKeys', keys=list(value.keys()) )
                else:
                    val = value[pos]
                key = str(pos)
                jobs.append\
                    ( defer.maybeDeferred\
                        ( self.validator.validate
                        , context, val
                        ).addCallback\
                            ( schema_gotResult
                            , result
                            , key
                            , isList
                            , self.returnList
                            )\
                        .addErrback\
                            ( schema_gotError
                            , errorset
                            , key
                            )
                    )
        else:
            for (key, val) in value.items():

                jobs.append\
                    ( defer.maybeDeferred\
                        ( self.validator.validate
                        , context, val
                        ).addCallback\
                            ( schema_gotResult
                            , result
                            , key
                            , isList
                            , self.returnList
                            )\
                        .addErrback\
                            ( schema_gotError
                            , errorset
                            , key
                            )
                    )

        d = defer.Deferred()
        jobs = defer.DeferredList( jobs )
        jobs.addCallback\
            ( schema__on_value_done
            , d
            , self
            , value
            , result
            , errorset
            )

        return d


    def forEach__createContextChildren_on_value( self, context, value ):
        isList = isinstance( value, list) or isinstance(value, tuple) or isinstance(value, set)

        if not isList:
            if not isinstance(value, dict ):
                raise Invalid( value, self,'type' )

        if self.returnList:
            result = []
        else:
            result = {}
        errors = []

        # populate
        children = []
        if isList or self.numericKeys:
            context.setIndexFunc( lambda index: str(index) )

            for pos in range( len( value ) ):
                if not isList:
                    val = value.get(str(pos),MISSING)
                    if value.get(str(pos),MISSING) is MISSING:
                        context.setIndexFunc( None )
                        raise Invalid( value, self, 'numericKeys',keys=list(value.keys()))

                else:
                    val = value[ pos ]

                contextChild = context( str( pos ) )
                contextChild.validator = self.validator
                contextChild.__value__ = val
                children.append( contextChild )

        else:
            context.setIndexFunc( None )

            if self.returnList:
                raise Invalid( value, self, 'listType' )
            for (key,val) in value.items():
                contextChild = context( key )
                contextChild.validator = self.validator
                contextChild.__value__ = val
                children.append( contextChild )

        jobs = []
        #validate
        for childContext in children:
            jobs.append\
                ( childContext.validate()\
                    .addCallback\
                        ( schema_gotResult
                        , result
                        , childContext.key
                        , isList
                        , self.returnList
                        )\
                    .addErrback\
                        ( schema_gotError
                        , errors
                        , childContext.key
                        )
                )

        d = defer.Deferred()
        jobs = defer.DeferredList( jobs )
        jobs.addCallback\
            ( schema__createContextChildren_on_value_done
            , d
            , self
            , value
            , result
            , errors
            )

        return d



    @defer.inlineCallbacks
    def field_validate(self, context, value):
        fieldcontext = self.getField( context, self.path )

        if not self.useResult:
            result = fieldcontext.value

        else:
            try:
                result = yield fieldcontext.result
            except Invalid:
                result = PASS

        if self.validator is not None:
            if result is not PASS:
                result = yield defer.maybeDeferred\
                    ( self.validator.validate
                    , fieldcontext, result
                    )

        if self.writeToContext:
            fieldcontext.__result__ = result

        if self.copy:
            if result is PASS:
                defer.returnValue( value )

            defer.returnValue( result )

        defer.returnValue( value )

    from twisted.names import client
    from twisted.names.dns import Record_MX
    from twisted.names.error import DNSNameError
    from twisted.internet.defer import TimeoutError

    def mxLookup_gotResult(result, d, value, validator, context ):
        if isinstance( result, Failure ):
            if isinstance(result.value, TimeoutError):
                d.errback( Invalid( value, validator ) )
            elif not isinstance(result.value, DNSNameError):
                d.errback( result )
            else:
                d.errback( Invalid( value, validator ) )
            return

        (answers, auth, add) = result
        if not len(answers):
            d.errback( Invalid( value, validator ) )
        else:
            for record in answers:
                if isinstance(record.payload,Record_MX):
                    d.callback( value )
                    return

            d.errback( Invalid( value, validator ) )

    mxLookup_resolver = client.Resolver('/etc/resolv.conf')
    def mxLookup_on_value( self, context, value ):
        d = defer.Deferred()
        mxLookup_resolver.lookupMailExchange( value, [2,4,6,8,10] )\
            .addBoth( mxLookup_gotResult, d, value, self, context )

        return d


    Context.validate = context_validate
    Tag.validate = tag_validate
    Compose.valdate = compose_validate
    Tmp.validate = tmp_validate
    Item.validate = item_validate
    Not.validate = not_validate
    And.validate = and_validate
    Or.validate = or_validate
    Call.validate = call_validate
    Match.on_value = match_on_value
    If.validate = if_validate
    Schema._on_value = schema__on_value
    Schema._createContextChildren_on_value = schema__createContextChildren_on_value
    ForEach._on_value = forEach__on_value
    ForEach._createContextChildren_on_value = forEach__createContextChildren_on_value
    Field.validate = field_validate
    MXLookup.on_value = mxLookup_on_value

    monkeyPatch._isMonkeyPatched = True

from ..util import getArgSpec, getParameterNames



def validateDecorator_gotValidationResult\
        ( result
        , d
        , origArgs
        , origKwargs
        , method
        , varargs
        , keywords
        , shifted
        , onInvalid
        ):

    if isinstance( result, Failure ):
        if not isinstance(result.value, Invalid):
            d.errback( result )
        elif onInvalid is not None:
            try:
                result = onInvalid( result.value )
            except Exception as e:
                d.errback( e )
            else:
                d.callback( result )
        else:
            d.errback( result )

    else:
        origKwargs.update( result )
        
        resultArgs = origKwargs.pop( varargs, origArgs )
        resultArgs = [ origKwargs.pop(key) for key in shifted  ] + resultArgs
        
        if keywords is not False:
            origKwargs.update( origKwargs.pop( keywords ) )
        
        defer.maybeDeferred( method, *resultArgs, **origKwargs )\
            .chainDeferred( d )
   

def validateDecorator( validator, method, include, exclude, onInvalid, inlineCallbacks ):

    if include and exclude:
        raise SyntaxError("'include' and 'exclude' cannot be used at the same time")

    spec = getArgSpec( method )
    hasVarargs = spec.varargs is not None
    varargs =  spec.varargs or '*varargs'
    keywords = spec.keywords or False

    methodParameterNames = getParameterNames( method, skipSelf=False )

    skip = ()
    if exclude:
        skip = exclude
    if include:
        skip = set(methodParameterNames) - set(include)

    varargs    = varargs

    hasVarargs = spec.varargs not in skip and hasVarargs

    keywords   = keywords not in skip and keywords

    if inlineCallbacks:
        method = defer.inlineCallbacks( method )

    def __wrap( *fargs, **fkwargs):

        d = defer.Deferred()

        (fargs, fkwargs, shifted ) = varargs2kwargs( method, fargs, fkwargs, skipSelf=False )
        origKwargs = dict(fkwargs)

        if keywords is not False:
            restKwargs = dict(\
                ( key, fkwargs.pop(key))\
                    for key in list(fkwargs.keys()) if key not in methodParameterNames
                )
            fkwargs[ keywords ] = restKwargs

        if fargs or hasVarargs:
            fkwargs[ varargs ] = list(fargs)

        result = validator.context\
            ( dict( ( key, fkwargs[ key] ) for key in fkwargs if key not in skip )
            ).result

        result.addBoth( validateDecorator_gotValidationResult, d, fargs, origKwargs, method, varargs, keywords, shifted, onInvalid )

        return d

    return __wrap

def validate( validator, include=None, exclude=None, onInvalid=None, inlineCallbacks=False ):
    def __createDecorator( method ):
        return validateDecorator( validator, method, include, exclude, onInvalid, inlineCallbacks)
    return __createDecorator




import sys
_python3 = sys.version_info[0]>=3

class Invalid(Exception):
    context = None

    def __init__(self, value, _validator=None, _key='fail', **kwargs):
        if _validator is not None:
            self.validator = _validator
        self.value = value
        self.data = {'key': _key, 'extra': kwargs}

    @property
    def key(self):
        return self.data['key']

    @key.setter
    def key(self, value):
        self.data['key'] = value

    @property
    def message(self):
        return self.data.get('message',None)

    @property
    def extra(self):
        return self.data['extra']

    def __repr__(self):
        return 'Invalid(%s, %s)' % (self.value, self.key)

    def __str__(self):
        text = self.__unicode__()
        if not _python3:
            return text.encode('utf8')
        return text

    def __unicode__(self):
        if self.context is not None and self.message is not None:
            return self.context.root.errorFormatter( self.context, self )
        else:
            return self.__repr__()




from setuptools import setup, find_packages

setup\
    ( name = 'kanone'
    , version = '0.4.1'
    , description = 'a validation library'
    , long_description = open('README.txt').read()
    , author = 'don`catnip'
    , author_email = 'don dot t at pan1 dot cc'
    , url = 'http://github.com/doncatnip/kanone'
    , classifiers =\
        [ "Development Status :: 4 - Beta"
        , "Topic :: Software Development :: Libraries :: Python Modules"
        , "License :: Public Domain"
        , "Programming Language :: Python :: 2.7"
        , "Programming Language :: Python :: 3"
        , 'Intended Audience :: Developers'
        ]
    , license = 'Unlicense'
    , keywords = 'validation library form twisted stateful'

    , packages = find_packages('src')
    , package_dir = {'':'src'}
    , install_requires = [ "zope.interface" ]
    #, namespace_packages = ['kanone' ]
    , include_package_data = True
    )



from ..lib import Context as __Context__, MISSING, PASS, inherit, Parameterized
from ..error import Invalid

import logging
from functools import reduce

log = logging.getLogger(__name__)

#### Basic stuff

def messages( **_messages ):
    def decorate( klass ):
        klassMessages = dict(getattr(klass,'__messages__', {} ))
        klassMessages.update( _messages )

        setattr( klass, '__messages__', klassMessages )
        return klass

    return decorate


def validator2parameter( hostValidator, name, param, force=False ):
    paramWrapper =  getattr(hostValidator,'__paramWrapper__',None)
    if not paramWrapper:
        class ParamWrapper:
            def __init__( self, context, value ):
                self.__context = context
                self.__value = value

        hostValidator.__paramWrapper__ =\
            paramWrapper =\
            ParamWrapper

        validate = hostValidator.validate
        def wrapped( context, value ):
            context.params = paramWrapper( context, value )
            return validate( context, value )

        hostValidator.validate = wrapped

    if isinstance(param, ValidatorBase ):
        hostValidator.__paramValidators__.append( param )
        prop = property(lambda self: param.validate( self.__context, self.__value ))
    else:
        if force:
            raise SyntaxError('Parameter has to be a validator')
        prop = param
        
    setattr( paramWrapper, name, prop )


# validators without messages and changeable parameters should derive from this
class ValidatorBase(object):

    def __new__( cls, *args, **kwargs ):
        self = object.__new__( cls )

        preValidators =\
            getattr(cls,'__pre_validate__',[])
        postValidators =\
            getattr(cls,'__post_validate__',[])

        if preValidators or preValidators:
            self.__init__( *args, **kwargs )
            self = And( *preValidators + [ self ] + postValidators )

        return self

    def appendSubValidators( self, subValidators ):
        pass

    def __and__( self, other ):
        if isinstance(other, And):
            other.validators.insert(0, self)
            return other
        return And( self, other )

    def __or__( self, other ):
        if isinstance(other, Or):
            other.validators.insert(0,self)
            return other
        return Or( self, other )

    def __invert__( self ):
        return Not( self )

    def tag( self, tagName, enabled=True ):
        return Tag( self, tagName, enabled )

    def context( self, value=MISSING ):
        return __Context__( self, value )

    # just a passthrough for convinience
    def __call__( self ):
        return self

@messages\
    ( fail='Validation failed'
    , missing= 'Please provide a value'
    , blank='Field cannot be empty'
    )
@inherit\
    ( '__messages__'
    )
class Validator( Parameterized, ValidatorBase ):

    def __init__( self, *args, **kwargs ):
        Parameterized.__init__( self, *args, **kwargs )

    def validate( self, context, value ):
        if value is MISSING:
            return self.on_missing( context )
        elif value is None or (value == ''):
            return self.on_blank( context, value )

        return self.on_value( context, value )

    def messages( self, **messages):
        self.__messages__ = dict( self.__messages__ )
        self.__messages__.update( messages )
        return self

    def on_value( self, context, value ):
        return value

    def on_missing(self, context):
        raise Invalid( '', self, 'missing' )

    def on_blank(self, context, value ):
        raise Invalid( value, self, 'blank' )


class Tag( ValidatorBase ):

    _ID = 0

    def __init__( self, validator, tagName, enabled=True):
        if isinstance( validator, Tag ):
            raise SyntaxError('%s is not taggable' % validator.__class__.__name__ )

        self.validator = validator
        self.tagName = tagName
        self.enabled = enabled
        self.tagID = Tag._ID
        Tag._ID += 1

    def appendSubValidators( self, subValidators ):
        self.validator.appendSubValidators( subValidators )
        subValidators.append( self.validator )

    def validate( self, context, value ):
        validator = context.root.taggedValidators.get(self.tagID, None)
        if validator is None:
            validator = self.enabled and self.validator

        if not validator:
            return value

        try:
            return validator.validate( context, value )
        except Invalid as e:
            if e.validator is validator or getattr(e,'composer',None) is validator:
                e.tagName = self.tagName
            raise e

def _setParsedKeywordArg( tagKwargs, key, value ):
    tagPath = key.split('_',1)

    theKwargs = tagKwargs.get(tagPath[0], None)
    if theKwargs is None:
        theKwargs = tagKwargs[tagPath[0]] = {}
   
    if len(tagPath)==2:
        theKwargs[tagPath[1]] = value
    else:
        # maybe TODO: we could use these values as varargs
        # but for now, just set it to raise an error if the tag
        # doesnt exist
        theKwargs['_none__'] = None

def _parseTaggedKeywords( kwargs, alias ):

    tagKwargs = {}

    for (key,value) in kwargs.items():
        if key.startswith('_'):
            continue

        if alias and (key in alias):
            if isinstance( alias[key], tuple ):
                for realKey in alias[key]:
                    _setParsedKeywordArg( tagKwargs, realKey, value )
            elif hasattr( alias[key], '__call__' ):
                realKwargs = alias[key]( key, value )
                for (realKey, realValue) in realKwargs.items():
                    _setParsedKeywordArg( tagKwargs, realKey, realValue )
            else:
                _setParsedKeywordArg( tagKwargs, alias[key], value )
            continue

        _setParsedKeywordArg( tagKwargs, key, value )

    return tagKwargs

@inherit\
    ( '__paramAlias__'
    , '__messageAlias__'
    , 'tagIDs'
    , 'taggedValidators'
    , 'currentTaggedValidators'
    , 'validator'
    )
class Compose( Validator ):

    """
    usage:
    UserName = Compose\
          ( String.convert().tag('string') & Len(min=5).tag('len')
          ).paramAlias( len='len_min' ).messageAlias(type='string_type')

    someUserName = UserName( len=8 ).message( type='Must be a string !' )

    An alias can be: single tag, list of tags, function receiving alias, value
    and returning dict { realtag:value, .. }
    """

    # stuff defined here will be inherited by children of this Validator


    __paramAlias__ = None
    __messageAlias__ = None

    taggedValidators = {}

    def setArguments( self, validator ):
        self.validator = validator
        self.taggedValidators = {}
        self.currentTaggedValidators = {}
        self.tagIDs = {}
        subValidators = [ self.validator ]
        self.validator.appendSubValidators( subValidators )

        for validator in subValidators:
            if isinstance( validator, Tag):
                tagName = validator.tagName
                tagID = validator.tagID
                enabled = validator.enabled
                validator = validator.validator
                if not tagName in self.tagIDs:
                    self.tagIDs[tagName] = []
                self.taggedValidators[ tagID ] = validator
                self.currentTaggedValidators[ tagID ] = enabled and validator
                self.tagIDs[tagName].append(tagID)

        if not self.tagIDs:
            raise SyntaxError('No tags found.')

    def setParameters( self, **kwargs):
        taggedKwargs = _parseTaggedKeywords( kwargs, self.__paramAlias__ )

        notFound = []
        self.__clonedTaggedValidators = []

        if not self.__isRoot__:
            self.taggedValidators = dict( self.taggedValidators )
            self.currentTaggedValidators = dict( self.currentTaggedValidators )

        for (tagName, args) in taggedKwargs.items():
            tagIDs = self.tagIDs.get( tagName, None )

            if tagIDs is None:
                notFound.append( tagName )
            else:
                enabled = args.pop('enabled',None)

                for tagID in tagIDs:
                    if enabled is not None:
                        self.currentTaggedValidators[ tagID ] =\
                            enabled and self.taggedValidators[ tagID ] or False
                    else:
                        if args:
                            self.taggedValidators[ tagID ] =\
                                validator =\
                                self.taggedValidators[ tagID ](**args)
                            self.__clonedTaggedValidators.append( tagID )
                        else:
                            validator = self.taggedValidators[ tagID ]

                        self.currentTaggedValidators[ tagID ]\
                            = validator

        if notFound:
            raise SyntaxError('setParameters: Tags %s not found' % str(notFound))

    def validate( self, context, value ):
        tmpTags = context.root.taggedValidators
        context.root.taggedValidators = self.currentTaggedValidators
        try:
            return self.validator.validate( context, value )
        except Invalid as e:
            if hasattr(e,'tagName'):
                e.realkey = "%s_%s" % (e.tagName, getattr(e,'realkey',e.key))
                e.composer = self
                del e.tagName
            raise e
        finally:
            context.root.taggedValidators = tmpTags

    def messages( self, **kwargs ):
        taggedKwargs = _parseTaggedKeywords( kwargs, self.__messageAlias__ )

        notFound = []

        for (tagName,args) in taggedKwargs.items():
            tagIDs = self.tagIDs.get(tagName,None)
            if tagIDs is None:
                notFound.append( tagName )
            else:
                for tagID in tagIDs:
                    if tagID not in self.__clonedTaggedValidators:
                        self.taggedValidators[tagID]\
                            = taggedValidator \
                            = self.taggedValidators[tagID]()
                        if self.currentTaggedValidators[tagID] is not False:
                            self.currentTaggedValidators[tagID] = taggedValidator

                        self.__clonedTaggedValidators.append ( tagID )
                    else:
                        taggedValidator = self.taggedValidators[tagID]

                    taggedValidator.messages( **args )

        if notFound:
            raise SyntaxError('messages: Tags %s not found' % str(notFound))

        return self

    def messageAlias( self, **alias ):
        self.__messageAlias__ = alias
        return self

    def paramAlias( self, **alias ):
        self.__paramAlias__ = alias
        return self


class Tmp( ValidatorBase ):

    def __init__( self, validator, raiseError=True ):
        self.validator = validator
        self.raiseError = raiseError

    def appendSubValidators( self, subValidators ):
        subValidators.append( self.validator )
        self.validator.appendSubValidators( subValidators )
        return subValidators

    def validate( self, context, value ):
        try:
            self.validator.validate( context, value )
        except Invalid:
            if self.raiseError:
                raise

        return value

@messages\
    ( type='Unsupported type, must be list-like or dict'
    , notFound='Item %(key)s not found'
    )
class Item( Validator ):

    def setParameters( self, key=1, validator=None, alter=None ):
        if not validator and alter:
            raise SyntaxError('alter can only be set to True, if a validator is given')

        self.key = key
        self.validator = validator
        if alter is not None:
            self.alter = alter
        else:
            self.alter = validator is not None

    def appendSubValidators( self, subValidators ):
        subValidators.append( self.validator )
        self.validator.appendSubValidators( subValidators )

    def validate( self, context, value ):
        try:
            val = value[ self.key ]
        except TypeError:
            raise Invalid( value, self, 'type' )
        except (KeyError, IndexError):
            raise Invalid( value, self, 'notFound', key=self.key )
        else:
            if self.validator is not None:
                val = self.validator.validate( context, val )

                if self.alter:
                    value[self.key] = val
                return value
            else:
                return val


class If( ValidatorBase ):

    def __init__( self, criterion, _then, _else=None):
        self.criterion = criterion
        self._then = _then
        self._else = _else

    def appendSubValidators( self, subValidators ):
        self.criterion.appendSubValidators( subValidators )
        self._then.appendSubValidators( subValidators )
        subValidators.append( self.criterion )
        subValidators.append( self._then )

        if self._else:
            self._else.appendSubValidators( subValidators )
            subValidators.append( self._else )

    def validate( self, context, value ):
        try:
            value = self.criterion.validate( context, value )
        except Invalid:
            if not self._else:
                raise
            value = self._else.validate( context, value)
        else:
            value = self._then.validate( context, value )

        return value

class Pass( Validator ):

    def setParameters( self, default=PASS):
        self.default = PASS

    def validate( self, context, value ):
        if self.default is PASS:
            return value

        return self.default

@messages\
    ( fail='Field must not match criterion'
    )
@inherit\
    ( 'validator'
    )
class Not( Validator ):

    def setArguments(self, criterion):
        self.validator = criterion

    def appendSubValidators( self, subValidators):
        self.validator.appendSubValidators( subValidators )
        subValidators.append( self.validator )

    def validate(self, context, value ):
        try:
            self.validator.validate\
                ( context, value )
        except Invalid:
            return value

        raise Invalid( value, self )


class And( ValidatorBase ):

    def __init__( self, *validators ):
        assert len(validators)>=2
        self.validators = list(validators)

    def validate( self, context, value ):
        return reduce( lambda val, validator: validator.validate(context, val), self.validators, value)

    def appendSubValidators( self, subValidators):
        for validator in self.validators:
            validator.appendSubValidators( subValidators )
            subValidators.append( validator )

    def __and__( self, other ):
        self.validators.append( other )
        return self

class Or( ValidatorBase ):

    def __init__( self, *validators ):
        assert len(validators)>=2
        self.validators = list(validators)

    def appendSubValidators( self, subValidators):
        for validator in self.validators:
            validator.appendSubValidators( subValidators )
            subValidators.append( validator )

    def validate(self, context, value):
        lastError = None

        for validator in self.validators:
            try:
                return validator.validate( context, value )
            except Invalid as e:
                lastError = e

        if lastError is not None:
            raise lastError

        return value

    def __or__( self, other ):
        self.validators.append( other )
        return self


class Call( Validator ):

    def setParameters( self, func ):
        self.__func__ = func

    def validate( self, context, value ):
        try:
            return self.__func__( context, value )
        except Invalid as e:
            e.validator = self
            raise e

